"""Arrrgs starter"""
import asyncio
from logging import DEBUG
from argparse import Namespace
from typing import Any, Callable, List, Dict, Tuple, Union
from inspect import signature
from .parser import parser
from .log import get_logger

log = get_logger("arrrgs.run")

ContextPreparer = Callable[[Namespace, Any], Tuple[Namespace, Any]]

def run(ctx: Any = None, prepare: Union[ContextPreparer, None] = None, debug=False):
    """Runs application"""
    asyncio.run(async_run(ctx, prepare, debug))

async def async_run(ctx: Any = None, prepare: Union[ContextPreparer, None] = None, debug=False):
    """Runs application asynchronously"""
    parsed_args = parser.parse_args()
    if debug:
        log.setLevel(DEBUG)
        log.info("Running in debug mode")
    log.info(parsed_args)
    if prepare is not None:
        log.info("Running 'prepare' callback")
        (prepared_args, prepared_context) = await _run_safe(prepare, parsed_args, ctx)
        # Updating values
        parsed_args = prepared_args
        ctx = prepared_context
    handler: Callable
    if parsed_args.cmd is None:
        if parsed_args.no_command_handler is None:
            log.info("no_command handler is not found. Printing help")
            parser.print_help()
            return
        handler = parsed_args.no_command_handler
    else:
        handler = parsed_args.func
    await _run_safe(handler, parsed_args, ctx)

def _prepare_args(handler: Callable, parsed_args: Namespace, ctx: Any) -> List[str]:
    sig = signature(handler)
    params_count = len(sig.parameters)
    args = []
    if params_count >= 1:
        args.append(parsed_args)
    if params_count >= 2:
        args.append(ctx)
    return args

async def _run_safe(handler: Callable, parsed_args: Dict[str, Any], ctx: Any):
    args = _prepare_args(handler, parsed_args, ctx)
    return await _run_async_safe(handler, args)

async def _run_async_safe(handler: Callable, args: List[any]):
    if asyncio.iscoroutinefunction(handler):
        return await handler(*args)
    return handler(*args)



"""Arrrgs log utils"""
from logging import getLogger, Formatter, StreamHandler, Logger

def get_logger(name: str) -> Logger:
    """Return a logger with the specified name."""
    log = getLogger(name)
    stream_handler = StreamHandler()
    stream_handler.setFormatter(Formatter(
        "%(asctime)s - [%(levelname)s] - %(name)s - %(message)s"
    ))
    log.addHandler(stream_handler)
    return log



"""Arrrgs command"""
from .parser import command_subparsers, parser

def command(*args, name:str=None, parent=command_subparsers):
    """Decorator to define a new command"""
    def decorator(func):
        if name is not None:
            command_name = name
        else:
            command_name = func.__name__
        cmd_parser = parent.add_parser(command_name, description=func.__doc__)
        for arg in list(args):
            cmd_parser.add_argument(*arg[0], **arg[1])
        cmd_parser.set_defaults(func=func)
    return decorator

def no_command():
    """Decorator to define a command absence handler"""
    def decorator(func):
        parser.set_defaults(no_command_handler=func)
    return decorator



"""Arrrgs arguments"""
from typing import Tuple, Dict, Any
from .parser import parser

def arg(*name_or_flags: Tuple[str, ...], **kwargs: Dict[str, Any]) -> None:
    """Convenience function to properly format arguments
    to pass to the subcommand decorator."""
    return (list(name_or_flags), kwargs)

def global_args(*args: Tuple[str, ...]) -> None:
    """Adds global arguments to app"""
    for argument in list(args):
        parser.add_argument(*argument[0], **argument[1])



"""Arrrgs setup script"""
import setuptools

with open(".version", "r", encoding="utf-8") as fh:
    VERSION = ''.join(fh.read().split())

with open("README.md", "r", encoding="utf-8") as fh:
    README = fh.read()

setuptools.setup(
    name="arrrgs",
    version=VERSION,
    author="Mikhael Khrustik",
    description="The library for easily writing feature-rich Python scripts",
    long_description=README,
    long_description_content_type="text/markdown",
    packages=[
        'arrrgs'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License"
    ],
    python_requires='>=3.7',
    package_dir={'':'.'},
)
